/*
 *  Copyright (C) 2004-2012 Savoir-Faire Linux Inc.
 *  Author: Yan Morin <yan.morin@savoirfairelinux.com>
 *  Author : Laurielle Lea <laurielle.lea@savoirfairelinux.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  If you modify this program, or any covered work, by linking or
 *  combining it with the OpenSSL project's OpenSSL library (or a
 *  modified version of that library), containing parts covered by the
 *  terms of the OpenSSL or SSLeay licenses, Savoir-Faire Linux Inc.
 *  grants you additional permission to convey the resulting work.
 *  Corresponding Source for a non-source form of such a combination
 *  shall include the source code for the parts of OpenSSL used as well
 *  as that of the covered work.
 */
#include "call.h"
#include "manager.h"
#include "audio/mainbuffer.h"
#include "history/historyitem.h"
#include "scoped_lock.h"

Call::Call(const std::string& id, Call::CallType type)
    : callMutex_()
    , localIPAddress_("")
    , localAudioPort_(0)
    , localVideoPort_(0)
    , id_(id)
    , confID_()
    , type_(type)
    , connectionState_(Call::DISCONNECTED)
    , callState_(Call::INACTIVE)
    , isIPToIP_(false)
    , peerNumber_()
    , displayName_()
    , timestamp_start_(0)
    , timestamp_stop_(0)
{
    pthread_mutex_init(&callMutex_, NULL);
    time(&timestamp_start_);
}

Call::~Call()
{
    pthread_mutex_destroy(&callMutex_);
}

void
Call::setConnectionState(ConnectionState state)
{
    sfl::ScopedLock m(callMutex_);
    connectionState_ = state;
}

Call::ConnectionState
Call::getConnectionState()
{
    sfl::ScopedLock m(callMutex_);
    return connectionState_;
}

void
Call::setState(CallState state)
{
    sfl::ScopedLock m(callMutex_);
    callState_ = state;
}

Call::CallState
Call::getState()
{
    sfl::ScopedLock m(callMutex_);
    return callState_;
}

std::string
Call::getStateStr()
{
    switch (getState()) {
        case ACTIVE:
            switch (getConnectionState()) {
                case RINGING:
                    return isIncoming() ? "INCOMING" : "RINGING";
                case CONNECTED:
                default:
                    return "CURRENT";
            }

        case HOLD:
            return "HOLD";
        case BUSY:
            return "BUSY";
        case INACTIVE:

            switch (getConnectionState()) {
                case RINGING:
                    return isIncoming() ? "INCOMING" : "RINGING";
                case CONNECTED:
                    return "CURRENT";
                default:
                    return "INACTIVE";
            }

        case CONFERENCING:
            return "CONFERENCING";
        case REFUSED:
        case ERROR:
        default:
            return "FAILURE";
    }
}


std::string
Call::getLocalIp()
{
    sfl::ScopedLock m(callMutex_);
    return localIPAddress_;
}

unsigned int
Call::getLocalAudioPort()
{
    sfl::ScopedLock m(callMutex_);
    return localAudioPort_;
}

unsigned int
Call::getLocalVideoPort()
{
    sfl::ScopedLock m(callMutex_);
    return localVideoPort_;
}

bool
Call::setRecording()
{
    bool recordStatus = Recordable::recAudio_.isRecording();

    Recordable::recAudio_.setRecording();
    MainBuffer &mbuffer = Manager::instance().getMainBuffer();
    std::string process_id = Recordable::recorder_.getRecorderID();

    if (!recordStatus) {
        mbuffer.bindHalfDuplexOut(process_id, id_);
        mbuffer.bindHalfDuplexOut(process_id, MainBuffer::DEFAULT_ID);

        Recordable::recorder_.start();
    } else {
        mbuffer.unBindHalfDuplexOut(process_id, id_);
        mbuffer.unBindHalfDuplexOut(process_id, MainBuffer::DEFAULT_ID);
    }

    Manager::instance().getMainBuffer().dumpInfo();

    return recordStatus;
}

void Call::time_stop()
{
    time(&timestamp_stop_);
}

std::string Call::getTypeStr() const
{
    switch (type_) {
        case INCOMING:
            return "incoming";
        case OUTGOING:
            return "outgoing";
        case MISSED:
            return "missed";
        default:
            return "";
    }
}

namespace {
    std::string timestamp_to_string(const time_t &timestamp)
    {
        std::stringstream time_str;
        time_str << timestamp;
        return time_str.str();
    }
}

std::map<std::string, std::string> Call::createHistoryEntry() const
{
    using sfl::HistoryItem;
    std::map<std::string, std::string> result;

    result[HistoryItem::ACCOUNT_ID_KEY] = Manager::instance().getAccountFromCall(id_);
    result[HistoryItem::CONFID_KEY] = confID_;
    result[HistoryItem::CALLID_KEY] = id_;
    result[HistoryItem::DISPLAY_NAME_KEY] = displayName_;
    result[HistoryItem::PEER_NUMBER_KEY] = peerNumber_;
    result[HistoryItem::RECORDING_PATH_KEY] = recAudio_.fileExists() ? getFilename() : "";
    result[HistoryItem::TIMESTAMP_START_KEY] = timestamp_to_string(timestamp_start_);
    result[HistoryItem::TIMESTAMP_STOP_KEY] = timestamp_to_string(timestamp_stop_);
    if (connectionState_ == RINGING)
        result[HistoryItem::STATE_KEY] = HistoryItem::MISSED_STRING;
    else
        result[HistoryItem::STATE_KEY] = getTypeStr();
    return result;
}

std::map<std::string, std::string>
Call::getDetails()
{
    std::map<std::string, std::string> details;
    std::ostringstream type;
    type << type_;
    details["CALL_TYPE"] = type.str();
    details["PEER_NUMBER"] = peerNumber_;
    details["DISPLAY_NAME"] = displayName_;
    details["CALL_STATE"] = getStateStr();
    details["CONF_ID"] = confID_;
    details["TIMESTAMP_START"] = timestamp_to_string(timestamp_start_);
    return details;
}

std::map<std::string, std::string>
Call::getNullDetails()
{
    std::map<std::string, std::string> details;
    details["ACCOUNTID"] = "";
    details["PEER_NUMBER"] = "Unknown";
    details["PEER_NAME"] = "Unknown";
    details["DISPLAY_NAME"] = "Unknown";
    details["CALL_STATE"] = "UNKNOWN";
    details["CALL_TYPE"] = "0";
    return details;
}
